import pytest
from unittest.mock import MagicMock


@pytest.fixture
def mock_cb_manager_patch(monkeypatch):
    """
    Fixture que devuelve una instancia de cbManager con varias partes mockeadas:
    - session: para no hacer llamadas HTTP reales.
    - métodos internos como get_entities, send_batch, __send_batch, etc.
    """

    def _patch(
        ruta_importacion: str,
        get_entities_return_value=None,
        get_entities_side_effect=None,
        get_entities_page_return_value=None,
        get_entities_page_side_effect=None,
        send_batch_return_value=True,
        send_batch_side_effect=None,
        delete_entities_return_value=None,
        delete_entities_side_effect=None,
    ):
        mock_cb = MagicMock()
        mock_cb.session = MagicMock()

        mock_cb.get_entities.return_value = get_entities_return_value
        if get_entities_side_effect:
            mock_cb.get_entities.side_effect = get_entities_side_effect

        mock_cb.get_entities_page.return_value = get_entities_page_return_value
        if get_entities_page_side_effect:
            mock_cb.get_entities_page.side_effect = get_entities_page_side_effect

        mock_cb.send_batch.return_value = send_batch_return_value
        if send_batch_side_effect:
            mock_cb.send_batch.side_effect = send_batch_side_effect

        mock_cb.delete_entities.return_value = delete_entities_return_value
        if delete_entities_side_effect:
            mock_cb.delete_entities.side_effect = delete_entities_side_effect

        # Inyectar mock en la ruta indicada
        monkeypatch.setattr(ruta_importacion, lambda *args, **kwargs: mock_cb)
        return mock_cb

    return _patch



import logging
import sys
import traceback

FAILED_EXEC = 1
SUCCESS_EXEC = 0

class FlowControl:
    '''Clase que implementa métodos que se usan para controlar
    el flujo de la ETL, por ejemplo para terminar la ejecución,
    para permitir que continue pero que envie un correo al final
    o para indicar que la ejecución ha sido exitosa'''

    _instance = None
    flow_state = None
    log = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._initialize_flow_control()
        return cls._instance

    @staticmethod
    def _initialize_flow_control():
        '''Método estático para inicializar los atributos de FlowControl'''
        FlowControl.flow_state = SUCCESS_EXEC
        FlowControl.log = logging.getLogger()

    def handle_error(self, cause: str, fatal: bool = False) -> None:
        '''Logea el error recibido, lo añade al correo que se va a enviar y
        si es un error fatal termina la ejecución de la ETL'''
        self.flow_state = FAILED_EXEC
        self.log.error(cause)
        self.log.error(f'Excepción: {traceback.format_exc(limit=1)}')

        if fatal:
            self.end_exec()

    def end_exec(self):
        '''Termina la ejecución de la ETL retornando 1 en caso de fallo y 0
        en caso se exito'''
        if self.flow_state == FAILED_EXEC:
            self.log.info('Ejecución Fallida, ETL KO')
        else:
            self.log.info('Ejecución Exitosa, ETL OK')

        return sys.exit(self.flow_state)



import logging
import time
from functools import wraps


def log_timer(func):
    '''Logea el tiempo que tarda en ejecutarse una función'''

    @wraps(func)
    def log_func(*args, **kwargs):
        start_time = time.time()
        result = func(*args)
        exc_time = time.time() - start_time
        log = logging.getLogger()
        log.info("La función {} se ejecuto en {} segundos".format(
            func.__name__, exc_time))

        return result
    return log_func



from pandas import DataFrame


CARACTERES_PROHIBIDOS_CONTEXT_BROKER = {">": "[", "<": "]", "‘": "´", "“": "´", "=": "-", ";": ",", "(": "[", ")": "]", "'": "´", '"': "´"}


def reemplazar_caracteres_prohibidos_context_broker(df: DataFrame) -> DataFrame:
    df = df.copy()
    
    for columna in df.columns:
        if df[columna].dtype == object:
            for caracter, sustituto in CARACTERES_PROHIBIDOS_CONTEXT_BROKER.items():
                df[columna] = df[columna].str.replace(caracter, sustituto, regex=False)
    return df



from unittest.mock import MagicMock
import pytest


@pytest.fixture
def mock_oracle_patch(monkeypatch):
    """
    Fixture que mockea OracleConnector para evitar conexiones reales a Oracle.

    - Mockea los métodos execute_query, call_procedure y call_function.
    - Cada uno puede devolver una lista de mocks con atributo ._mapping (para simular filas).
    - Se puede simular error con side_effect.
    - Uso similar a mock_sqlserver_patch.

    Uso:
        def test_execute_query(mock_oracle_patch):
            mock_oracle = mock_oracle_patch(
                "modulo.OracleConnector",
                execute_query_return=[{"columna": "valor"}]
            )
            resultado = mock_oracle.execute_query()
            assert resultado[0]._mapping["columna"] == "valor"

        def test_call_function_error(mock_oracle_patch):
            mock_oracle = mock_oracle_patch(
                "modulo.OracleConnector",
                call_function_side_effect=Exception("Error en función")
            )
            with pytest.raises(Exception, match="Error en función"):
                mock_oracle.call_function("func_error")
    """

    def _patch(
        ruta_importacion: str,
        execute_query_return=None,
        execute_query_side_effect=None,
        execute_many_query_return=None,
        execute_many_query_side_effect=None,
        call_procedure_return=None,
        call_procedure_side_effect=None,
        call_function_return=None,
        call_function_side_effect=None,
    ):
        mock_connector_instance = MagicMock()

        if execute_query_return is not None:
            rows = []
            for row_dict in execute_query_return:
                row_mock = MagicMock()
                row_mock._mapping = row_dict
                rows.append(row_mock)
            mock_connector_instance.execute_query.return_value = rows
        if execute_query_side_effect is not None:
            mock_connector_instance.execute_query.side_effect = execute_query_side_effect

        if execute_many_query_return is not None:
            rows = []
            for row_dict in execute_many_query_return:
                row_mock = MagicMock()
                row_mock._mapping = row_dict
                rows.append(row_mock)
            mock_connector_instance.execute_many_query.return_value = rows
        if execute_many_query_side_effect is not None:
            mock_connector_instance.execute_many_query.side_effect = execute_many_query_side_effect

        if call_procedure_return is not None:
            rows = []
            for row_dict in call_procedure_return:
                row_mock = MagicMock()
                row_mock._mapping = row_dict
                rows.append(row_mock)
            mock_connector_instance.call_procedure.return_value = rows
        if call_procedure_side_effect is not None:
            mock_connector_instance.call_procedure.side_effect = call_procedure_side_effect

        if call_function_return is not None:
            mock_connector_instance.call_function.return_value = call_function_return
        if call_function_side_effect is not None:
            mock_connector_instance.call_function.side_effect = call_function_side_effect

        mock_constructor = MagicMock(return_value=mock_connector_instance)
        monkeypatch.setattr(ruta_importacion, mock_constructor)

        return mock_connector_instance

    return _patch




from sqlalchemy import text
import logging
from typing import Any, List, Sequence, Tuple
import pandas as pd
from sqlalchemy import Row, TextClause, create_engine
from sqlalchemy.orm import sessionmaker
from vlcishared.utils.interfaces import ConnectionInterface
from sqlalchemy.engine import Connection


class PostgresConnector(ConnectionInterface):
    _instance = None

    @classmethod
    def instance(cls):
        if not cls._instance:
            raise Exception("PostgresConnector no ha sido inicializado aún.")
        return cls._instance

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(PostgresConnector, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, host: str, port: str, database: str, user: str, password: str):
        if self._initialized:
            return  # Ya fue inicializado, evitar reiniciar
        self.log = logging.getLogger()
        self.db_name = database
        self.connection_string = f"postgresql://{user}:{password}@{host}:{port}/{database}"
        self.connect()
        self._initialized = True

    def connect(self):
        """Función que se conecta a la base de datos definida en el constructor"""
        self.engine = create_engine(self.connection_string)
        self.session_maker = sessionmaker(bind=self.engine)
        self.session = self.session_maker()
        self.log.info(f"Conectado a {self.db_name}")

    def close(self):
        """Cierra la conexión con la base de datos"""
        if self.session:
            self.session.close()
        if self.engine:
            self.engine.dispose()
        self.log.info(f"Desconectado de {self.db_name}.")

    def call_procedure(self, procedure_name: str, *params: Any):
        """
        Llama a un procedimiento almacenado en la base de datos.

        Args:
            procedure_name (str): Nombre del procedimiento almacenado.
            *params (Any): Parámetros posicionales que se pasan al procedimiento.
                Se asignan automáticamente a placeholders (:p0, :p1, ...).

        Raises:
            Exception: Si la ejecución del procedimiento falla.
        """
        try:
            param_placeholders, param_dict = self._create_placeholder_params(params)
            sql = text(f"CALL {procedure_name}({param_placeholders})")
            self.session.execute(sql, param_dict)
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            self.log.error(f"Fallo llamando al procedimiento {procedure_name}: {e}")
            raise e

    def call_function(self, function_name: str, *params: Any) -> Sequence[Row[Any]]:
        """
        Llama a una función almacenada en la base de datos y devuelve su resultado.

        Args:
            function_name (str): Nombre de la función almacenada.
            *params (Any): Parámetros posicionales que se pasan al procedimiento.
                Se asignan automáticamente a placeholders (:p0, :p1, ...).

        Returns:
            Sequence[Row[Any]]: Lista de filas devueltas por la función.
                - Si la función retorna un escalar: se obtiene un único Row con una sola columna.
                - Si la función retorna un setof: se obtiene una lista de Row con múltiples registros.

        Raises:
            Exception: Si la ejecución de la función falla.
        """
        try:
            param_placeholders, param_dict = self._create_placeholder_params(params)
            sql = text(f"SELECT {function_name}({param_placeholders})")
            result = self.session.execute(sql, param_dict)
            self.session.commit()
            return result.fetchall()
        except Exception as e:
            self.session.rollback()
            self.log.error(f"Fallo llamando a la función {function_name}: {e}")
            raise e

    def _create_placeholder_params(self, params: List[Any]):
        placeholders = ", ".join([f":p{i}" for i in range(len(params))])
        param_dict = {f"p{i}": param for i, param in enumerate(params)}
        return placeholders, param_dict

    def execute_query(self, query: str | TextClause, params: dict = None, commit: bool = True, conn: Connection = None):
        """
        Ejecuta una consulta SQL, con o sin commit.

        Args:
            query (str | TextClause): Consulta SQL a ejecutar.
            params (dict, optional): Parámetros para la query.
            commit (bool, optional):
                - Si True: se abre una transacción con `engine.begin()` y se hace commit automáticamente al salir del contexto.
                - Si False: se abre una conexión simple con `engine.connect()` (sin transacción implícita, sin commit). Útil para SELECTs.
            conn (Connection, optional):
                Conexión ya existente a reutilizar. En este caso, no se abre ni se cierra nada dentro de esta función
                y el control de commit/rollback queda fuera.

        Returns:
            Result: El resultado de la ejecución de la query.

        Raises:
            Exception: Si la ejecución de la consulta falla.
        """
        try:
            query_obj = text(query) if isinstance(query, str) else query
            if conn:
                return conn.execute(query_obj, params or {})
            else:
                execution_context = self.engine.begin() if commit else self.engine.connect()
                with execution_context as conn:
                    result = conn.execute(query_obj, params or {})
                return result
        except Exception as e:
            self.log.error(f"Error al ejecutar la consulta: {e}")
            raise

    def execute_multiple_queries(self, queries_with_params: List[Tuple[str, dict]], conn: Connection = None):
        """
        Ejecuta múltiples consultas SQL distintas dentro de una única transacción. Si alguna falla, se hace rollback automático.
        Cada consulta es ejecutada una vez con su conjunto específico de parámetros.

        Args:
            queries_with_params (List[Tuple[str, dict]]):
                Lista de tuplas donde cada tupla contiene:
                - query (str): Consulta SQL parametrizada.
                - params (dict): Diccionario con valores para los parámetros de la consulta.
            conn (Connection, optional):
                Conexión ya existente a reutilizar. En este caso, no se abre ni se cierra nada dentro de esta función
                y el control de commit/rollback queda fuera.

        Raises:
            Exception: Si la ejecución de la consulta falla.
        """
        try:
            if conn:
                for query, params in queries_with_params:
                    conn.execute(text(query), params)
            else:
                with self.engine.begin() as trans:
                    for query, params in queries_with_params:
                        trans.execute(text(query), params)
            self.log.info("Todas las queries se ejecutaron correctamente.")
        except Exception as e:
            self.log.error(f"Error al ejecutar las queries: {e}")
            raise

    def insert_many_rows(self, query: str | TextClause, df: pd.DataFrame, conn: Connection = None):
        """
        Ejecuta una única consulta parametrizada en batch para insertar múltiples filas.
        La consulta SQL debe estar preparada para recibir múltiples sets de parámetros (una fila por set),
        típicamente un INSERT con placeholders nombrados.

        Args:
            query (str): Consulta SQL parametrizada.
                Debe usar placeholders con nombre, por ejemplo: "INSERT INTO tabla (col1, col2) VALUES (:col1, :col2)"
            df (pd.DataFrame): DataFrame que contiene los datos a insertar.
                Cada fila se convertirá en un diccionario de parámetros para la consulta SQL.
            conn (Connection, optional):
                Conexión ya existente a reutilizar. En este caso, no se abre ni se cierra nada dentro de esta función
                y el control de commit/rollback queda fuera.

        Raises:
            Exception: Si la ejecución de la consulta falla.

        """
        try:
            if df.empty:
                self.log.info("DataFrame vacío, no se inserta nada")
                return

            query_obj = text(query) if isinstance(query, str) else query
            params_list = df.to_dict(orient="records")

            if conn:
                conn.execute(query_obj, params_list)
            else:
                with self.engine.begin() as trans:
                    trans.execute(query_obj, params_list)
            self.log.info(f"Ejecutadas {len(df)} filas con la consulta parametrizada.")
        except Exception as e:
            self.log.error(f"Error al insertar múltiples registros: {e}")
            raise



from abc import ABC, abstractmethod
from vlcishared.tokens.token_provider import TokenProvider


class ArcGISTokenProvider(TokenProvider, ABC):
    """
    Clase base abstracta para proveedores de tokens de ArcGIS.
    Define la estructura y delega en las subclases la generación
    del payload concreto.
    """

    def __init__(self, url_token_provider, username, password, format, expiration):
        super().__init__(url_token_provider)
        self.username = username
        self.password = password
        self.format = format
        self.expiration = expiration

    @abstractmethod
    def get_payload(self):
        """
        Cada subclase debe implementar este método para generar el
        payload específico de la solicitud de token.
        """
        pass

    def obtener_token(self):
        """
        Realiza la solicitud del token al servidor ArcGIS usando las credenciales
        y parámetros configurados. Asigna el token recibido a self.token.

        Lanza:
            ValueError: Si la respuesta no contiene un token válido.
        """
        payload = self.get_payload()
        respuesta = self._solicitar_nuevo_token(data=payload)
        token = respuesta.get("token")
        if token is None:
            raise ValueError("No se ha podido obtener el token.")
        self.token = token


import pytest
from unittest.mock import MagicMock


from unittest.mock import MagicMock
import pytest
from vlcishared.csv.generador_csv import GeneradorCSV

@pytest.fixture
def mock_generador_csv_patch(monkeypatch):
    """
    Fixture reutilizable que mockea la clase GeneradorCSV para evitar escribir archivos reales.

    - Mockea el método `exportar`.
    - Permite inspeccionar llamadas y rutas generadas.
    - Permite inyectar side_effect para simular excepciones.

    Uso:
        def test_xxx(mock_generador_csv_patch):
            mock_csv = mock_generador_csv_patch("gestion_csv.gestor_csv.GeneradorCSV")
            ruta = mock_csv.exportar("archivo.csv", [["dato1"]], ["col1"])
            assert mock_csv.exportar.call_count == 1
    """

    def _patch(ruta_importacion: str, carpeta_exportacion: str, exportar_side_effect=None):
        instancia_real = GeneradorCSV(carpeta_exportacion) 
        generador_mock = MagicMock(spec=GeneradorCSV)
        rutas_csv = []

        def side_effect_exportar(nombre_fichero, filas, cabecera=None):
            ruta = instancia_real.exportar(nombre_fichero, filas, cabecera)
            rutas_csv.append(ruta)

            if exportar_side_effect:
                return exportar_side_effect(nombre_fichero, filas, cabecera)
            return ruta

        generador_mock.exportar.side_effect = side_effect_exportar
        generador_mock.exportar.rutas_csv = rutas_csv

        monkeypatch.setattr(ruta_importacion, lambda *args, **kwargs: generador_mock)

        return generador_mock

    return _patch



from sqlalchemy import text

from vlcishared.env_variables.secrets import get_secret


def copiar_tablas(connection, esquema_origen, esquema_destino):
    """
    Copia todas las tablas del esquema original al de test. Se ejecuta una vez por sesión.
    Usa la misma conexión que el resto de los tests.
    """
    tablas = (
        connection.execute(
            text(
                """
            SELECT tablename
            FROM pg_tables
            WHERE schemaname = :esquema
            """
            ),
            {"esquema": esquema_origen},
        )
        .scalars()
        .all()
    )

    for tabla in tablas:
        connection.execute(
            text(
                f"""
            DO $$
            BEGIN
                IF NOT EXISTS (
                    SELECT 1 FROM pg_tables WHERE tablename = '{tabla}' AND schemaname = '{esquema_destino}'
                ) THEN
                    EXECUTE 'CREATE TABLE {esquema_destino}.{tabla} (LIKE {esquema_origen}.{tabla} INCLUDING ALL)';
                END IF;
            END
            $$;
        """
            )
        )

    connection.commit()


def copiar_funciones_y_procedimientos(connection, esquema_origen, esquema_destino):

    funciones = (
        connection.execute(
            text(
                """
            SELECT p.oid
            FROM pg_proc p
            JOIN pg_namespace n ON p.pronamespace = n.oid
            WHERE n.nspname = :esquema
            """
            ),
            {"esquema": esquema_origen},
        )
        .scalars()
        .all()
    )

    for oid in funciones:
        definicion = connection.execute(text("SELECT pg_get_functiondef(:oid)"), {"oid": oid}).scalar()

        if not definicion:
            continue

        definicion = definicion.replace(f"{esquema_origen}.", f"{esquema_destino}.")

        try:
            connection.execute(text(definicion))
        except Exception as e:
            print(f"Error creando función/procedimiento {oid}: {e}")

    connection.commit()


def borrar_tablas(connection, esquema):
    tablas = (
        connection.execute(
            text(
                """
                SELECT tablename
                FROM pg_tables
                WHERE schemaname = :esquema
                """
            ),
            {"esquema": esquema},
        )
        .scalars()
        .all()
    )
    for tabla in tablas:
        connection.execute(text(f"DROP TABLE IF EXISTS {esquema}.{tabla} CASCADE"))
    connection.commit()


def borrar_funciones_y_procedimientos(connection, esquema):
    funciones = connection.execute(
        text(
            """
            SELECT
                p.proname,
                pg_get_function_identity_arguments(p.oid) AS args,
                p.prokind
            FROM pg_proc p
            JOIN pg_namespace n ON p.pronamespace = n.oid
            WHERE n.nspname = :esquema
            """
        ),
        {"esquema": esquema},
    ).fetchall()

    for nombre, args, tipo in funciones:
        drop_type = "PROCEDURE" if tipo == "p" else "FUNCTION"
        try:
            connection.execute(text(f'DROP {drop_type} IF EXISTS {esquema}."{nombre}"({args}) CASCADE'))
        except Exception as e:
            print(f"Error borrando {drop_type.lower()} {nombre}({args}): {e}")
            connection.rollback()

    connection.commit()



from datetime import datetime
import os
import re
import zipfile

def ordenar_ficheros_por_fecha_en_nombre(formato_fecha, ficheros):
    """
    Ordena una lista de nombres de ficheros según la fecha contenida en su nombre.

    Args:
        formato_fecha (str): Formato en el que está escrita la fecha en el nombre del fichero (por ejemplo: '%Y%m%d', '%Y_%m_%d').
        ficheros (list of str): Lista de nombres de ficheros a ordenar.

    Returns:
        list of str: Lista de nombres de ficheros ordenados cronológicamente según la fecha extraída.
    """
    archivos_con_fecha = []
    for fichero in ficheros:
        fecha = extraer_fecha_desde_nombre_fichero(formato_fecha, fichero)
        archivos_con_fecha.append((fecha, fichero))

    archivos_con_fecha.sort(key=lambda x: x[0])
    return [nombre for _, nombre in archivos_con_fecha]

def extraer_fecha_desde_nombre_fichero(formato_fecha, nombre_fichero):
    """
    Extrae la fecha del nombre de un fichero usando un patrón regex y un formato específico.

    Args:
        formato_fecha (str): Formato en el que está escrita la fecha en el nombre del fichero (por ejemplo: '%Y%m%d', '%Y_%m_%d').
        nombre_fichero (str): Nombre del fichero del que extraer la fecha.

    Returns:
        datetime: Objeto datetime con la fecha extraída.

    Raises:
        ValueError: Si no se encuentra una fecha válida en el nombre del fichero o el formato no coincide.
    """
    patron_fecha_str = formato_a_patron_regex(formato_fecha)
    patron_fecha_compilado = re.compile(patron_fecha_str)
    match = patron_fecha_compilado.search(nombre_fichero)
    if not match:
        raise ValueError(f"El fichero {nombre_fichero} no contiene una fecha válida")

    fecha_str = match.group(0)
    return datetime.strptime(fecha_str, formato_fecha)

def formato_a_patron_regex(formato_fecha: str) -> str:
    r"""
    Convierte un formato de fecha de datetime a un patrón regex.
    
    Args:
        formato_fecha (str): como '%Y_%m_%d' o '%d-%m-%Y'
    
    Returns:
        str: patrón regex equivalente, ej: r"\d{4}_\d{2}_\d{2}"
    """
    mapping = {
        "%Y": r"\d{4}",
        "%y": r"\d{2}",
        "%m": r"\d{2}",
        "%d": r"\d{2}",
    }

    patron = formato_fecha
    for token, regex_equivalente in mapping.items():
        patron = patron.replace(token, regex_equivalente)

    patron = re.sub(r"(?<!\\)([^\w\\{}])", r"\\\1", patron)
    return patron

def comprimir_fichero_en_zip(ruta_fichero: str) -> str:
    """
    Comprime un fichero cualquiera en un archivo ZIP en la misma carpeta.

    Args:
        ruta_fichero (str): Ruta completa del fichero que se desea comprimir.

    Returns:
        str: Ruta completa del archivo ZIP creado que contiene el fichero original.
    """
    nombre_base = os.path.splitext(os.path.basename(ruta_fichero))[0]

    ruta_zip = os.path.join(os.path.dirname(ruta_fichero), f"{nombre_base}.zip")

    with zipfile.ZipFile(ruta_zip, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(ruta_fichero, arcname=os.path.basename(ruta_fichero))

    return ruta_zip



from vlcishared.tokens.arcgis_sigval_token_provider_config import ArcGISSigvalTokenProviderConfig
from vlcishared.tokens.arcgis_token_provider import ArcGISTokenProvider


class ArcGISSigvalTokenProvider(ArcGISTokenProvider):
    """
    Proveedor de tokens para ArcGIS que usa 'client' en el payload.
    """

    def __init__(self, config: ArcGISSigvalTokenProviderConfig):
        super().__init__(config.url_token_provider, config.username, config.password, config.format, config.expiration)
        self.client = config.client

    def get_payload(self):
        return {
            "username": self.username,
            "password": self.password,
            "f": self.format,
            "client": self.client,
            "expiration": self.expiration,
        }
    


from unittest.mock import MagicMock, Mock

import pytest
from requests import Response
from requests.exceptions import ConnectionError
from tc_etl_lib import exceptions


@pytest.fixture
def mock_iota_manager_patch(monkeypatch):
    """
    Fixture que mockea iotaManager para evitar conexiones HTTP reales.

    - Mockea los métodos send_http y send_batch_http
    - Permite simular respuestas exitosas y errores
    - Simula el comportamiento de requests.Session
    - Permite testear funciones que usan iotaManager sin hacer llamadas HTTP reales

    Uso básico:
        def test_send_success(mock_iota_manager_patch):
            mock_iota = mock_iota_manager_patch(
                "modulo.iotaManager",
                send_http_return=True
            )
            resultado = mock_iota.send_http({"temperatura": 25})
            assert resultado is True
    """

    def _patch(
        ruta_clase_mockear: str,
        retornar_constructor=False,
        send_http_return=None,
        send_http_side_effect=None,
        send_batch_http_return=None,
        send_batch_http_side_effect=None,
        endpoint="http://mock-endpoint",
        device_id="mock_device",
        api_key="mock_api_key",
        sleep_send_batch=0,
        timeout=10,
        post_retry_connect=3,
        post_retry_backoff_factor=20,
    ):
        """
        Función interna para configurar el mock de iotaManager.

        Args:
            ruta_clase_mockear: Ruta donde se importa iotaManager (ej: "modulo.iotaManager")
            send_http_return: Valor de retorno para send_http (por defecto True)
            send_http_side_effect: Excepción a lanzar en send_http
            send_batch_http_return: Valor de retorno para send_batch_http (por defecto True)
            send_batch_http_side_effect: Excepción a lanzar en send_batch_http
            endpoint, device_id, api_key, etc.: Parámetros del constructor
        """

        mock_iota_instance = MagicMock()

        mock_iota_instance.endpoint = endpoint
        mock_iota_instance.device_id = device_id
        mock_iota_instance.api_key = api_key
        mock_iota_instance.sleep_send_batch = sleep_send_batch
        mock_iota_instance.timeout = timeout
        mock_iota_instance.post_retry_connect = post_retry_connect
        mock_iota_instance.post_retry_backoff_factor = post_retry_backoff_factor

        mock_session = MagicMock()
        mock_iota_instance.session = mock_session

        mock_iota_instance.send_http.return_value = send_http_return or True
        mock_iota_instance.send_batch_http.return_value = send_batch_http_return or True

        if send_http_side_effect:
            mock_iota_instance.send_http.side_effect = send_http_side_effect

        if send_batch_http_side_effect:
            mock_iota_instance.send_batch_http.side_effect = send_batch_http_side_effect

        mock_constructor = MagicMock(return_value=mock_iota_instance)
        monkeypatch.setattr(ruta_clase_mockear, mock_constructor)

        if retornar_constructor:
            return mock_constructor, mock_iota_instance

        return mock_iota_instance

    return _patch


@pytest.fixture
def mock_iota_manager_with_response_patch(monkeypatch):
    """
    Fixture avanzada que mockea iotaManager con control detallado sobre las respuestas HTTP.

    Permite simular respuestas HTTP específicas, códigos de estado, y errores de conexión.

    Uso:
        def test_http_response_codes(mock_iota_manager_with_response_patch):
            mock_iota = mock_iota_manager_with_response_patch(
                "modulo.iotaManager",
                http_status_code=500,
                response_json={"error": "Internal Server Error"}
            )
            with pytest.raises(Exception):  # FetchError se lanzará
                mock_iota.send_http({"data": "test"})
    """

    def _patch(ruta_clase_mockear: str, http_status_code=200, response_json=None, connection_error=False, **kwargs):
        mock_iota_instance = MagicMock()

        for attr, value in kwargs.items():
            setattr(mock_iota_instance, attr, value)

        mock_response = Mock(spec=Response, status_code=http_status_code, **{"json.return_value": response_json or {}})

        mock_session = MagicMock()
        mock_iota_instance.session = mock_session

        if connection_error:
            mock_session.post.side_effect = ConnectionError("Connection failed")
        else:
            mock_session.post.return_value = mock_response

        def mock_send_http(data):
            if not isinstance(data, dict):
                raise TypeError("The 'data' parameter should be a dictionary with key-value pairs.")

            if connection_error:
                raise ConnectionError("Connection failed")

            if http_status_code == 200:
                return True
            else:
                raise exceptions.FetchError(
                    response=mock_response,
                    method="POST",
                    url=mock_iota_instance.endpoint,
                    params={
                        "i": mock_iota_instance.device_id,
                        "k": mock_iota_instance.api_key,
                    },
                    headers={"Content-Type": "application/json"},
                )

        mock_iota_instance.send_http.side_effect = mock_send_http

        mock_constructor = MagicMock(return_value=mock_iota_instance)
        monkeypatch.setattr(ruta_clase_mockear, mock_constructor)

        return mock_iota_instance

    return _patch



import ast
import logging

from requests import head
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class ClienteHTTP:
    """
    Cliente para realizar llamadas HTTP reutilizables basado en requests.Session.
    Permite descargar archivos y está diseñado para ser fácilmente ampliable.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, headers=None, timeout=10, verify_ssl=True):
        if self._initialized:
            return
        self.log = logging.getLogger()
        self.headers = headers or {}
        self.timeout = timeout
        self.verify_ssl = verify_ssl

        self.session = requests.Session()
        self.session.headers.update(self.headers)

        self._initialized = True

    def get(self, url, params=None, headers=None):
        params = params or {}
        request_headers = self.session.headers.copy()
        request_headers.update(headers or {})
        try:
            response = self.session.get(url, params=params, headers=request_headers, timeout=self.timeout, verify=self.verify_ssl)
            response.raise_for_status()
            return self._procesar_respuesta(response)
        except requests.RequestException as e:
            self.log.error(f"Error HTTP en GET a {url}: {e}")
            raise

    def post(self, url, *, data=None, files=None, headers=None):
        data = data or {}
        request_headers = self.session.headers.copy()
        request_headers.update(headers or {})
        try:
            response = self.session.post(url, data=data, files=files, headers=request_headers, timeout=self.timeout, verify=self.verify_ssl)
            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            self.log.error(f"Error HTTP en POST a {url}: {e}")
            raise

    def post_con_reintentos(self, url, *, data=None, files=None, headers=None, reintentos=3, delay=3):
        """
        Realiza un POST con reintentos automáticos en caso de fallo temporal.
        reintentos: número máximo de intentos (default: 3).
        delay: backoff_factor -> tiempo de espera entre reintentos (ej: 3 → 3s, 6s, 12s).
        """
        data = data or {}
        request_headers = self.session.headers.copy()
        request_headers.update(headers or {})

        session_tmp = requests.Session()
        session_tmp.headers.update(request_headers)

        retry_strategy = Retry(total=reintentos, backoff_factor=delay, status_forcelist=[408, 429, 500, 502, 503, 504, 507], allowed_methods=["POST"])
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session_tmp.mount("http://", adapter)
        session_tmp.mount("https://", adapter)

        try:
            response = session_tmp.post(url, data=data, files=files, headers=request_headers, timeout=self.timeout, verify=self.verify_ssl)
            response.raise_for_status()
            return self._procesar_respuesta(response)
        except requests.RequestException as e:
            self.log.error(f"Error HTTP en POST con reintentos a {url}: {e}")
            raise
        finally:
            session_tmp.close()

    def _procesar_respuesta(self, response):
        """
        Intenta parsear JSON si el servidor lo indica, de lo contrario devuelve texto.
        """
        content_type = response.headers.get("Content-Type", "").lower()
        if "application/json" in content_type:
            return response.json()
        else:
            return response.text

    def descargar_archivo_get(self, url, destino):
        """
        Descarga un archivo desde la URL dada y lo guarda en la ruta 'destino'.
        """
        try:
            response = self.session.get(
                url, 
                stream=True, 
                timeout=self.timeout, 
                verify=self.verify_ssl
            )
            response.raise_for_status()
            return self._guardar_respuesta_en_archivo(response, destino, url)
        
        except requests.RequestException as e:
            self.log.error(f"Error HTTP al descargar {url}: {e}")
            raise


    def descargar_archivo_post(self, url, destino, headers=None, body=None):
        """
        Descarga un archivo desde la URL usando HTTP POST y lo guarda en la ruta 'destino'.
        """
        request_headers = self.session.headers.copy()
        headers_dict = ast.literal_eval(headers) if headers else {}
        request_headers.update(headers_dict)
        
        try:
            response = self.session.post(
                url,
                data=body,
                headers=request_headers,
                stream=True,
                timeout=self.timeout,
                verify=self.verify_ssl
            )
            response.raise_for_status()
            return self._guardar_respuesta_en_archivo(response, destino, url)
        
        except requests.RequestException as e:
            self.log.error(f"Error HTTP al descargar {url}: {e}")
            raise



    def cerrar(self):
        """
        Cierra la sesión HTTP y libera recursos.
        """
        if self.session:
            self.session.close()
        self.log.info("Sesión HTTP cerrada correctamente.")

    def _guardar_respuesta_en_archivo(self, respuesta, destino, url):
        """
        Guarda el contenido de una respuesta HTTP en un archivo.
        """
        try:
            with open(destino, "wb") as f:
                for chunk in respuesta.iter_content(chunk_size=8192):
                    f.write(chunk)
            
            self.log.info(f"Archivo descargado correctamente desde {url} a {destino}")
            return True
        
        except Exception as e:
            self.log.error(f"Error al guardar archivo desde {url}: {e}")
            raise


from datetime import datetime
import fnmatch
import os
import shutil
from unittest.mock import MagicMock

import pytz

import pytest

from vlcishared.ftp.ftp import FTPClient
from vlcishared.tests.fixtures.ftp.sftp import patch_delete_side_effect, patch_upload_side_effect


@pytest.fixture
def mock_ftp_patch(monkeypatch):
    """
    Fixture reutilizable que mockea la clase FTPClient para evitar conexiones reales con FTP.

    - Mockea los métodos para evitar efectos reales.
    - Requiere pasar la ruta donde se usa FTPClient (para monkeypatching dinámico).

    Parámetros:
    - ruta_importacion (str): Ruta completa donde se importa `FTPClient` (ej. "mi_paquete.mi_modulo.FTPClient").
    - list_return (list): Lista de archivos que devolverá `list()`.
    - list_side_effect (callable|None): Función a ejecutar en `list()`.
    - list_matching_files_return (list): Lista de archivos que devolverá `list_matching_files()`.
    - list_matching_files_side_effect (callable|None): Función a ejecutar en `list_matching_files()`.
    - upload_side_effect (callable|None): Función a ejecutar en `upload()`.
    - delete_side_effect (callable|None): Función a ejecutar en `delete()`.
    - execute_with_connection_side_effect (callable|None): Función a ejecutar en `execute_with_connection()`.

    Uso:
        def test_xxx(mock_ftp_patch):
            mock_ftp = mock_ftp_patch("modulo.donde.importa.FTPClient", list_return=["file1.csv"], ...)
    """

    def _patch(
        ruta_importacion: str,
        list_return=[],
        list_side_effect=None,
        list_matching_files_return=[],
        list_matching_files_side_effect=None,
        download_side_effect=patch_ftp_download_side_effect,
        upload_side_effect=patch_upload_side_effect,
        delete_side_effect=patch_delete_side_effect,
        execute_with_connection_side_effect=None,
    ):
        ftp_mock = MagicMock(spec=FTPClient)
        ftp_mock.connect.return_value = None
        ftp_mock.close.return_value = None

        ftp_mock.list.return_value = list_return
        if list_side_effect:
            ftp_mock.list.side_effect = list_side_effect

        ftp_mock.list_matching_files.return_value = list_matching_files_return
        if list_matching_files_side_effect:
            ftp_mock.list_matching_files.side_effect = list_matching_files_side_effect

        ftp_mock.download.side_effect = lambda remote_path, local_path, pattern, min_date=None: download_side_effect(
            ftp_mock, remote_path, local_path, pattern, min_date
        )
        ftp_mock.upload.side_effect = upload_side_effect
        ftp_mock.delete.side_effect = delete_side_effect

        if execute_with_connection_side_effect:
            ftp_mock.execute_with_connection.side_effect = execute_with_connection_side_effect
        else:
            ftp_mock.execute_with_connection.side_effect = lambda func, *args, **kwargs: patch_ftp_execute_with_connection_side_effect(
                ftp_mock, func, *args, **kwargs
            )

        monkeypatch.setattr(ruta_importacion, lambda *args, **kwargs: ftp_mock)
        return ftp_mock

    return _patch


def patch_ftp_execute_with_connection_side_effect(ftp_mock, func, *args, **kwargs):
    """
    Simula execute_with_connection ejecutando la función pasada sin manejar conexiones reales.
    """
    return func(*args, **kwargs)


def patch_ftp_list_matching_files_side_effect(directorio, patron):
    """
    Simula list_matching_files listando archivos del directorio que coincidan con el patrón.
    """
    archivos = os.listdir(directorio)
    archivos_coincidentes = fnmatch.filter(archivos, patron)
    return archivos_coincidentes


def patch_ftp_download_side_effect(self, remote_path, local_path, pattern, min_date=None):
    """
    Simula la descarga de archivos desde remote_path a local_path,
    filtrando por patrón regex y copiando los archivos locales que coincidan.
    """
    archivos = os.listdir(remote_path)
    matching_files = fnmatch.filter(archivos, pattern)
    for archivo in matching_files:
        origen = os.path.join(remote_path, archivo)
        destino = os.path.join(local_path, archivo)

        fecha_mod = datetime.fromtimestamp(os.path.getmtime(origen), pytz.UTC)

        # Filtrar por min_date si se proporciona
        if min_date and fecha_mod <= min_date:
            continue

        shutil.copyfile(origen, destino)



import logging

from pymongo import MongoClient, errors


class MongoDBConnector():
    def __init__(self, host: str, port: str, auth_source: str, user: str, password: str):
        self.uri = (
            f"mongodb://{user}:{password}@"
            f"{host}:{port}/"
            f"{auth_source}?authSource={auth_source}"
        )
        self.host = host
        self.port = port
        self.auth_source = auth_source
        self.user = user
        self.password = password
        self.client = None
        self.log = logging.getLogger()

    def get_db_connection(self):
        """Función que se conecta a la base de datos de MongoDB
            definida en el constructor"""
        try:
            self.log.info('Conectando a MongoDB')
            self.client = MongoClient(self.uri)
            return self.client.get_database()
        except errors.ConnectionFailure as e:
            raise ConnectionError(f"Error de conexión a MongoDB: {e}")

    def disconnect(self):
        if self.client:
            self.log.info('Desconectando de MongoDB')
            self.client.close()

    def insert_data(self, collection_name: str, data):
        """
        Function to insert data into a specified MongoDB collection.
        """

        db = self.get_db_connection()
        collection = db[collection_name]

        try:
            if isinstance(data, list):
                self.log.info(f'Insert multiple docs into {collection_name}')
                result = collection.insert_many(data)
                self.log.info(f'Inserted documents')
            else:
                self.log.info(f'Insert a single doc into {collection_name}')
                result = collection.insert_one(data)
                self.log.info(f'Inserted document ID: {result.inserted_id}')
        except errors.PyMongoError as e:
            self.log.error(f"Error while inserting data into MongoDB: {e}")
            raise
        finally:
            self.disconnect()



from vlcishared.flow.flow import FAILED_EXEC


def comprobar_handle_error(mock_flow_control, causa):
    mock_flow_control.handle_error.assert_called_once()
    kwargs = mock_flow_control.handle_error.call_args.kwargs
    assert causa in kwargs["cause"]


def comprobar_sys_exit_fallido(mock_sys_exit, llamadas_esperadas):
    assert mock_sys_exit.call_count == llamadas_esperadas
    for call in mock_sys_exit.call_args_list:
        assert call == ((FAILED_EXEC,), {})


import json
import os


class GeneradorPayload:
    def __init__(self, carpeta: str):
        self.carpeta = carpeta

    def generar_payload(self, nombre_archivo: str) -> list[dict]:
        ruta = os.path.join(self.carpeta, nombre_archivo)
        with open(ruta, "r", encoding="utf-8") as f:
            return json.load(f)


from datetime import datetime
import os
import re
import shutil
from unittest.mock import MagicMock

import pytz

import pytest

from vlcishared.sftp.sftp import SFTPClient


@pytest.fixture
def mock_sftp_patch(monkeypatch):
    """
    Fixture reutilizable que mockea la clase SFTPClient para evitar conexiones reales con SFTP.

    - Reutilizable en distintos módulos donde se importe SFTPClient.
    - Mockea los métodos: list, list_sorted_date_modification, download (con shutil.copyfile), move, upload, delete, sftp.rename y execute_with_connection para evitar efectos reales.
    - Devuelve una función que permite especificar el path del import y el retorno de list().
    - Requiere pasar la ruta donde se usa SFTPClient (para monkeypatching dinámico).

    Parámetros:
    - ruta_importacion (str): Ruta completa donde se importa `SFTPClient` (ej. "mi_paquete.mi_modulo.SFTPClient").
    - listdir_return (list): Lista de archivos que devolverá `os.listdir()`.
    - listdir_side_effect (callable|None): Función a ejecutar en `os.listdir()`.
    - list_matching_files_return (list): Lista de archivos que devolverá `list_matching_files()`.
    - list_matching_files_side_effect (callable|None): Función a ejecutar en `list_matching_files()`.
    - list_all_files_return (list): Lista de archivos que devolverá `list_all_files()`.
    - list_all_files_side_effect (callable|None): Función a ejecutar en `list_all_files()`.
    - list_sorted_date_modification_return (list): Resultado que devolverá `list_sorted_date_modification()`.
    - list_sorted_date_modification_side_effect (callable|None): Función a ejecutar en `list_sorted_date_modification()`.
    - get_side_effect (callable|None): Función a ejecutar en `get()`, por defecto usa `patch_get_side_effect`.
    - move_side_effect (callable|None): Función a ejecutar en `move()`.
    - upload_side_effect (callable|None): Función a ejecutar en `upload()`.
    - delete_side_effect (callable|None): Función a ejecutar en `delete()`.
    - sftp_rename_side_effect (callable|None): Función a ejecutar en `sftp.rename()`.
    - execute_with_connection_side_effect (callable|None): Función a ejecutar en `execute_with_connection()`.

    Uso:
        def test_xxx(mock_sftp_patch):
            mock_sftp = mock_sftp_patch("modulo.donde.importa.SFTPClient", list_return=["file1.csv"], ...)
    """

    def _patch(
        ruta_importacion: str,
        listdir_return=[],
        listdir_side_effect=None,
        list_matching_files_return=[],
        list_matching_files_side_effect=None,
        list_all_files_return=[],
        list_all_files_side_effect=None,
        list_sorted_date_modification_return=[],
        list_sorted_date_modification_side_effect=None,
        download_side_effect=patch_download_side_effect,
        move_side_effect=patch_move_side_effect,
        upload_side_effect=patch_upload_side_effect,
        delete_side_effect=patch_delete_side_effect,
        sftp_rename_side_effect=patch_sftp_rename_side_effect,
        execute_with_connection_side_effect=None,
    ):
        sftp_mock = MagicMock(spec=SFTPClient)
        sftp_mock.connect.return_value = None
        sftp_mock.close.return_value = None

        sftp_mock.sftp = MagicMock()
        sftp_mock.sftp.listdir.return_value = listdir_return
        if listdir_side_effect:
            sftp_mock.sftp.listdir.side_effect = listdir_side_effect

        sftp_mock.list_matching_files.return_value = list_matching_files_return
        if list_matching_files_side_effect:
            sftp_mock.list_matching_files.side_effect = list_matching_files_side_effect

        sftp_mock.list_all_files.return_value = list_all_files_return
        if list_all_files_side_effect:
            sftp_mock.list_all_files.side_effect = list_all_files_side_effect

        sftp_mock.list_sorted_date_modification.return_value = list_sorted_date_modification_return
        if list_sorted_date_modification_side_effect:
            sftp_mock.list_sorted_date_modification.side_effect = list_sorted_date_modification_side_effect

        sftp_mock.download.side_effect = lambda remote_path, local_path, pattern, min_date=None: download_side_effect(
            sftp_mock, remote_path, local_path, pattern, min_date
        )
        sftp_mock.move.side_effect = move_side_effect
        sftp_mock.upload.side_effect = upload_side_effect
        sftp_mock.delete.side_effect = delete_side_effect
        sftp_mock.sftp.rename.side_effect = sftp_rename_side_effect

        if execute_with_connection_side_effect:
            sftp_mock.execute_with_connection.side_effect = execute_with_connection_side_effect
        else:
            sftp_mock.execute_with_connection.side_effect = lambda func, *args, **kwargs: patch_execute_with_connection_side_effect(sftp_mock, func, *args, **kwargs)

        monkeypatch.setattr(ruta_importacion, lambda *args, **kwargs: sftp_mock)
        return sftp_mock

    return _patch


def patch_execute_with_connection_side_effect(sftp_mock, func, *args, **kwargs):
    """
    Simula execute_with_connection ejecutando la función pasada sin manejar conexiones reales.
    """
    return func(*args, **kwargs)


def patch_list_matching_files_side_effect(directorio, patron):
    """
    Simula list_matching_files listando archivos del directorio que coincidan con el patrón.
    """
    archivos = os.listdir(directorio)
    archivos_coincidentes = [f for f in archivos if re.fullmatch(patron, f)]
    return archivos_coincidentes


def patch_list_all_files_side_effect(directorio):
    return os.listdir(directorio)


def patch_download_side_effect(self, remote_path, local_path, pattern, min_date=None):
    """
    Simula la descarga de archivos desde remote_path a local_path,
    filtrando por patrón regex y copiando los archivos locales que coincidan.
    """
    archivos_remotos = self.sftp.listdir()
    matching_files = [f for f in archivos_remotos if re.fullmatch(pattern, f)]
    for archivo in matching_files:
        origen = os.path.join(remote_path, archivo)
        destino = os.path.join(local_path, archivo)

        fecha_mod = datetime.fromtimestamp(os.path.getmtime(origen), pytz.UTC)

        # Filtrar por min_date si se proporciona
        if min_date and fecha_mod <= min_date:
            continue

        shutil.copyfile(origen, destino)


def patch_move_side_effect(origen, destino, archivo):
    """
    Simula el mover de un archivo SFTP moviéndolo desde 'origen' a 'destino'.
    """
    shutil.move(os.path.join(origen, archivo), os.path.join(destino, archivo))


def patch_upload_side_effect(origen, destino):
    """
    Simula la subida de un archivo SFTP copiándolo desde 'origen' a 'destino'.
    """
    shutil.copyfile(origen, destino)


def patch_delete_side_effect(archivo):
    """
    Simula la eliminación de un archivo SFTP eliminándolo en la ruta recibida.
    """
    os.remove(archivo)


def patch_sftp_rename_side_effect(nombre_antiguo, nombre_nuevo):
    """
    Simula el renombramiento de un fichero de SFTP.
    """
    os.rename(nombre_antiguo, nombre_nuevo)


from unittest.mock import MagicMock

import pytest
from sqlalchemy import create_engine, text

from vlcishared.env_variables.secrets import get_secret
from vlcishared.tests.fixtures.db.schema_utils import borrar_funciones_y_procedimientos, borrar_tablas, copiar_funciones_y_procedimientos, copiar_tablas

from contextlib import contextmanager


@pytest.fixture
def mock_postgres_patch(monkeypatch, db_transaction):
    """
    Fixture que mockea PostgresConnector para ejecutar queries sobre una base de datos real (SQLite/PostgreSQL de prueba).

    - Redirige métodos como connect, execute y execute_query al engine de pruebas.
    - Permite testear sin conectarse a una base de datos real de producción.
    - Requiere la ruta del import de PostgresConnector.

    Parámetros:
    - ruta_importacion (str): Ruta completa donde se importa `PostgresConnector` (ej. "mi_paquete.mi_modulo.PostgresConnector").
    - call_procedure_side_effect (callable|None): Lógica personalizada a ejecutar cuando se llama `call_procedure`.
    - call_function_side_effect (callable|None): Lógica personalizada a ejecutar cuando se llama `call_function`.
    - execute_query_side_effect (callable|None): Lógica personalizada a ejecutar cuando se llama `execute_query`.
    - execute_multiple_queries_side_effect (callable|None): Lógica personalizada a ejecutar cuando se llama `execute_multiple_queries`.
    - insert_many_rows_side_effect (callable|None): Lógica personalizada a ejecutar cuando se llama `insert_many_rows`.

    Uso:
        def test_xxx(mock_db_patch):
            mock_db = mock_db_patch("modulo.donde.importa.PostgresConnector")
            mock_db.execute.assert_called_once()
    """

    def _patch(
        target_path: str,
        call_procedure_side_effect=None,
        call_function_side_effect=None,
        execute_query_side_effect=None,
        execute_multiple_queries_side_effect=None,
        insert_many_rows_side_effect=None,
    ):
        mock_connector = MagicMock()

        mock_connector.connect.return_value = db_transaction
        mock_connector.close.return_value = None

        mock_connector.call_procedure.side_effect = call_procedure_side_effect or call_procedure_side_effect_default(db_transaction)
        mock_connector.call_function.side_effect = call_function_side_effect or call_function_side_effect_default(db_transaction)
        mock_connector.execute_query.side_effect = execute_query_side_effect or execute_query_side_effect_default(db_transaction)
        mock_connector.execute_multiple_queries.side_effect = execute_multiple_queries_side_effect or execute_multiple_queries_side_effect_default(
            db_transaction
        )
        mock_connector.insert_many_rows.side_effect = insert_many_rows_side_effect or insert_many_rows_side_effect_default(db_transaction)
        mock_connector.engine = db_transaction.engine

        # @contextmanager te envuelve esa función en un objeto que implementa __enter__ y __exit__
        # sirve para cuando en la ETL hacemos with PostgresConnector.instance().engine.begin as conn:
        @contextmanager
        def fake_begin():
            trans = db_transaction.begin_nested()
            try:
                yield db_transaction  # esto es lo que se asigna a `conn` dentro del with (__enter__)
                trans.commit()  # esto actua como el __exit__
            except Exception:
                trans.rollback()  # esto actua como el __exit__
                raise

        mock_connector.engine.begin = fake_begin

        monkeypatch.setattr(target_path, lambda *args, **kwargs: mock_connector)
        return mock_connector

    yield _patch


@pytest.fixture
def db_transaction(connection):
    transaction = connection.begin_nested()
    yield connection
    transaction.rollback()


@pytest.fixture(scope="session")
def connection():
    """
    Crea una conexión de SQLAlchemy contra una base de datos PostgreSQL de prueba.

    - Devuelve una conexión viva que se cierra al finalizar el test.
    - Usado por otros fixtures para ejecutar queries reales en un entorno aislado.
    """
    user = get_secret("GLOBAL_DATABASE_POSTGIS_LOGIN_TEST")
    password = get_secret("GLOBAL_DATABASE_POSTGIS_PASSWORD_TEST")
    port = get_secret("GLOBAL_DATABASE_POSTGIS_PORT_TEST")
    host = get_secret("GLOBAL_DATABASE_POSTGIS_HOST_TEST")
    database = get_secret("GLOBAL_DATABASE_POSTGIS_DATABASE_TEST")
    schema = get_secret("GLOBAL_DATABASE_POSTGIS_SCHEMA_TEST")
    url = f"postgresql+psycopg2://{user}:{password}@{host}:{port}/{database}"
    engine = create_engine(url)
    connection = engine.connect()
    connection.execute(text(f"SET search_path TO {schema}"))
    yield connection
    connection.close()
    engine.dispose()


@pytest.fixture(scope="session", autouse=True)
def setup_esquema_test_desde_vlci2(connection):
    """
    Fixture que se ejecuta una vez por sesión de pytest (scope="session") y automáticamente en todas las pruebas (autouse=True).

    - Copia las tablas y funciones/procedimientos del esquema 'vlci2' al esquema de test.
    - Al finalizar la sesión, elimina las tablas y funciones/procedimientos del esquema de test.

    Si se necesita preparar el esquema de test a partir de un origen distinto, se puede definir un nuevo fixture personalizado en el módulo correspondiente:
    @pytest.fixture(scope="session", autouse=True)
        def setup_esquema_xxx_desde_yyy(connection):
            copiar_esquema(connection, esquema_yyy, esquema_xxx)
            yield
            limpiar_esquema(connection, esquema_xxx)
    """
    esquema_vlci2 = get_secret("GLOBAL_DATABASE_POSTGIS_SCHEMA_VLCI2")
    esquema_test = get_secret("GLOBAL_DATABASE_POSTGIS_SCHEMA_TEST")
    copiar_esquema(connection, esquema_vlci2, esquema_test)
    yield
    limpiar_esquema(connection, esquema_test)


def copiar_esquema(connection, esquema_origen, esquema_destino):
    """
    Copia tablas y funciones/procedimientos del esquema origen al destino usando la misma conexión.
    """
    copiar_tablas(connection, esquema_origen, esquema_destino)
    copiar_funciones_y_procedimientos(connection, esquema_origen, esquema_destino)


def limpiar_esquema(connection, esquema):
    """
    Borra tablas y funciones/procedimientos del esquema indicado usando la misma conexión.
    """
    borrar_tablas(connection, esquema)
    borrar_funciones_y_procedimientos(connection, esquema)


def call_procedure_side_effect_default(transaction):
    def _call_procedure(procedure_name, *params):
        placeholders = ", ".join([f":p{i}" for i in range(len(params))])
        param_dict = {f"p{i}": param for i, param in enumerate(params)}
        sql = text(f"CALL {procedure_name}({placeholders})")
        transaction.execute(sql, param_dict)

    return with_rollback(transaction, _call_procedure)


def call_function_side_effect_default(transaction):
    def _call_procedure(procedure_name, *params):
        placeholders = ", ".join([f":p{i}" for i in range(len(params))])
        param_dict = {f"p{i}": param for i, param in enumerate(params)}
        result = transaction.execute(text(f"SELECT {procedure_name}({placeholders})"), param_dict)
        return result.fetchall()

    return with_rollback(transaction, _call_procedure)


def execute_query_side_effect_default(transaction):

    def _execute_query(query, params=None, commit=True, conn=None):
        query_obj = text(query) if isinstance(query, str) else query
        if conn:
            return conn.execute(query_obj, params or {})
        else:
            if commit:
                return transaction.execute(query_obj, params or {})
            else:
                # Savepoint para aislar la query
                trans = transaction.begin_nested()
                try:
                    result = transaction.execute(query_obj, params or {})
                finally:
                    trans.rollback()  # revierte solo esta query
                return result

    return with_rollback(transaction, _execute_query)


def execute_multiple_queries_side_effect_default(transaction):

    def _execute_multiple_queries(queries_with_params, conn=None):
        if conn:
            for query, params in queries_with_params:
                conn.execute(text(query), params)
        else:
            trans = transaction.begin_nested()
            try:
                for query, params in queries_with_params:
                    transaction.execute(text(query), params)
            except Exception:
                trans.rollback()

    return with_rollback(transaction, _execute_multiple_queries)


def insert_many_rows_side_effect_default(transaction):

    def _insert_query_commit(query, df, conn=None):
        if df.empty:
            return

        query_obj = text(query) if isinstance(query, str) else query
        params_list = df.to_dict(orient="records")
        if conn:
            conn.execute(query_obj, params_list)
        else:
            transaction.execute(query_obj, params_list)

    return with_rollback(transaction, _insert_query_commit)


def with_rollback(transaction, func):
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            transaction.rollback()
            raise

    return wrapped


def execute_query_side_effect_reemplazo_esquema(transaction, esquema_original="vlci2", esquema_reemplazo="test_component"):  # noqa: F811
    def _execute(query, params=None, commit=True, conn=None):
        query_modificada = _reemplazar_query(query, esquema_original, esquema_reemplazo)
        funcion = execute_query_side_effect_default(transaction)
        return funcion(query_modificada, params, commit, conn)

    return _execute


def execute_multiple_queries_side_effect_reemplazo_esquema(transaction, esquema_original="vlci2", esquema_reemplazo="test_component"):  # noqa: F811
    def _execute(queries_with_params, conn=None):
        querys_modificadas = [(_reemplazar_query(query, esquema_original, esquema_reemplazo), params) for query, params in queries_with_params]
        funcion = execute_multiple_queries_side_effect_default(transaction)
        return funcion(querys_modificadas, conn)

    return _execute


def insert_many_rows_side_effect_reemplazo_esquema(transaction, esquema_original="vlci2", esquema_reemplazo="test_component"):  # noqa: F811
    def _execute(query, df, conn=None):
        query_modificada = _reemplazar_query(query, esquema_original, esquema_reemplazo)
        funcion = insert_many_rows_side_effect_default(transaction)
        return funcion(query_modificada, df, conn)

    return _execute


def _reemplazar_query(query, esquema_original, esquema_reemplazo):
    query_str = query if isinstance(query, str) else str(query)
    return query_str.replace(esquema_original, esquema_reemplazo)



import smtplib


class EmailServer:
    def __init__(self, smtp_server, smtp_port, smtp_user, smtp_password):
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.smtp_user = smtp_user
        self.smtp_password = smtp_password

    def get_smtp_connection(self):
        """Devuelve una conexión SMTP autenticada."""
        server = smtplib.SMTP(self.smtp_server, self.smtp_port)
        server.starttls()
        server.login(self.smtp_user, self.smtp_password)
        return server


from datetime import datetime

import pytest


@pytest.fixture
def mock_datetime_patch(monkeypatch):
    """
    Fixture que permite fijar datetime.now() a una fecha específica durante un test.

    - Mockea datetime.now() devolviendo siempre la misma fecha.
    - Requiere la ruta del import del módulo que usa datetime (para que surta efecto).

    Parámetros:
    - ruta_importacion (str): Ruta completa donde se importa `datetime` (ej. "mi_paquete.mi_modulo.datetime").
    - fecha (datetime.datetime): Fecha fija que se desea devolver cuando se llama a `now()` o `utcnow()`.

    Uso:
        def test_xxx(mock_datetime_patch):
            mock_datetime_patch(datetime(2025, 5, 29), "paquete.modulo.datetime")
    """

    def _patch(ruta_importacion, fecha):
        class MockDateTime(datetime):
            @classmethod
            def now(cls, tz=None):
                return fecha

            @classmethod
            def utcnow(cls):
                return fecha

        monkeypatch.setattr(ruta_importacion, MockDateTime)

    return _patch



class ArcGISGeoportalConfig:
    """
    Configuración base común para servicios ArcGIS.

    Parámetros comunes:
        url_geoportal: URL base del geoportal ArcGIS.
        endpoint_api_features: Endpoint API de features.
        true_curve: Booleano para mantener curvas en geometrías.
    """

    def __init__(self, url_geoportal, endpoint_api_features, format, true_curve):
        self.url_geoportal = url_geoportal
        self.endpoint_api_features = endpoint_api_features
        self.format = format
        self.true_curve = true_curve



import pytest

from vlcishared.tests.fixtures.db.postgresql import db_transaction, execute_query_side_effect_reemplazo_esquema, mock_postgres_patch  # noqa: F401


@pytest.fixture
def mock_postgres_gestion_fechas(mock_postgres_patch, db_transaction):  # noqa: F811
    mock = mock_postgres_patch(
        "gestion_fechas.repositorio.PostgresConnector.instance", execute_query_side_effect=execute_query_side_effect_reemplazo_esquema(db_transaction)
    )
    return mock
